'use strict';

var helpers = require('./helpers'),
    merge = helpers.merge,
    isFunc = helpers.isFunc,
    isString = helpers.isString,
    isPlainObject = helpers.isPlainObject;

/**
Contains the properties for a route
After attempting to match a uri to the Routes map

@class Route
@constructor
@private
**/
var Route = function Route(routes, uri) {
  this.uri = uri;
  this.matchedRoute = '';
  this.targets = [];
  this.actions = [];
  this.exits = [];
  this.options = {};
  this.notFound = null;
  this.fullMatch = false;

  this.matchOrNotFound(routes);

  this.uri = uri;
};

Route.prototype = {

  /**
  Attempt to match the uri, or call a notFound handler if nothing matches.
   @method matchOrNotFound
  @param {Object} routeLevel
  **/
  matchOrNotFound: function matchOrNotFound(routeLevel) {
    this.match(routeLevel);

    if (!this.fullMatch && this.notFound) {
      this.actions.push(this.notFound);
    }
  },

  /**
  Matches the uri from the routes map.
   @method match
  @param {Object} routeLevel
  **/
  match: function match(routeLevel) {
    var value, action, target;

    if (routeLevel.target) {
      this.targets.push(routeLevel.target);
    }

    if (this.targets.length) {
      target = this.targets[this.targets.length - 1];
    }

    if (routeLevel.notFound && target[routeLevel.notFound]) {
      this.notFound = {
        target: target,
        method: routeLevel.notFound
      };
    } else if (isFunc(routeLevel.notFound)) {
      this.notFound = this.anonymousAction(routeLevel.notFound);
    }

    for (var key in routeLevel) {
      if (routeLevel.hasOwnProperty(key)) {
        action = {
          target: target,
          method: null
        };
        value = routeLevel[key];

        if (this.isFragment(key) && this.isFragmentInURI(key)) {
          this.updateMatchedRoute(key);
          this.removeFragmentFromURI(key);

          if (this.isActionDescriptor(value)) {

            // Check that if this fragment is a namedParam,
            // we never override a regular fragment.
            if (!this.isNamedParam(key) || !action.method) {
              if (isString(value)) {
                action.method = value;
              } else if (isFunc(value)) {
                action = this.anonymousAction(value);
              } else {
                action.method = value.method;

                if (value.exit) {
                  this.exits.unshift({
                    method: value.exit,
                    target: routeLevel.target
                  });
                }

                if (value.options) {
                  this.mergeOptions(value.options);
                }
              }

              // Adding the action
              this.actions.push(action);

              if (key !== '/*') {
                this.fullMatch = true;
              }
            }
          } else if (value.hasOwnProperty('options')) {
            this.mergeOptions(value.options);
          }

          if (isPlainObject(value)) {
            // recurse
            this.match(value);
          }
        }
      }
    }
  },

  /**
  @method mergeOptions
  @param {Object} options
  **/
  mergeOptions: function mergeOptions(options) {
    this.options = merge(this.options, options);
  },

  /**
  appends the matched fragment to the matched route
   @method updateMatchedRoute
  @param {String} fragment
  **/
  updateMatchedRoute: function updateMatchedRoute(fragment) {
    if (fragment !== '/' && fragment !== '/*') {
      this.matchedRoute += fragment;
    }
  },

  /**
  removes matched fragments from the beginning of the uri
   @method removeFragmentFromURI
  @param {String} fragment
  **/
  removeFragmentFromURI: function removeFragmentFromURI(fragment) {
    var uri = this.uri,
        uriParts,
        subFrags;

    if (fragment !== '/' && fragment !== '/*') {
      if (this.includesNamedParam(fragment)) {
        uriParts = uri.split('/'), subFrags = fragment.split('/');

        subFrags.forEach(function (f, i) {
          if (f.indexOf(':') === 0) {
            uri = uri.replace('/' + uriParts[i], '');
          } else if (f) {
            uri = uri.replace('/' + f, '');
          }
        });
      } else {
        uri = uri.replace(fragment, '');
      }
    }

    this.uri = uri;
  },

  /**
  @method isFragmentInURI
  @param {Any} fragment
  @return {Boolean}
  **/
  isFragmentInURI: function isFragmentInURI(fragment) {
    var uri = this.uri,
        uriParts,
        subFrags;

    if (uri === '/' || uri === '') {
      return fragment === '/' || fragment === '/*';
    }

    if (fragment === '/') {
      return false;
    } else if (fragment === '/*') {
      return true;
    }
    // includes vs is named param
    else if (this.includesNamedParam(fragment)) {
        uriParts = uri.split('/'), subFrags = fragment.split('/');

        if (subFrags.length === 2) {
          return true;
        }

        return subFrags.map(function (f, i) {
          if (f.indexOf(':') === 0) {
            return true;
          } else {
            return uriParts[i].indexOf(f) !== -1;
          }
        }).reduce(function (a, b) {
          return a && b;
        });
      } else {
        return uri.indexOf(fragment) === 0;
      }
  },

  /**
  @method includesNamedParam
  @param {String} fragment
  @return {Boolean}
  **/
  includesNamedParam: function includesNamedParam(fragment) {
    return fragment.indexOf('/:') !== -1;
  },

  /**
  @method isFragment
  @param {Any} key
  @return {Boolean}
  **/
  isFragment: function isFragment(key) {
    return key.indexOf('/') === 0;
  },

  /**
  @method isActionDescriptor
  @param {Any} val
  @return {Boolean}
  **/
  isActionDescriptor: function isActionDescriptor(val) {
    return isString(val) || isFunc(val) || isPlainObject(val) && val.method;
  },

  /**
  @method isNamedParam
  @param {String} fragment
  @return {Boolean}
  **/
  isNamedParam: function isNamedParam(fragment) {
    return fragment.indexOf('/:') === 0;
  },

  /**
  Create an action Object from the given function.
   @method anonymousAction
  @param {Function} func the anonymous function to transform into an action.
  @return {Object} With a method key that maps to a generated String, and
  **/
  anonymousAction: function anonymousAction(func) {
    return {
      target: { lambda: func },
      method: 'lambda'
    };
  }
};

module.exports = Route;